package jp.riken.neurodata.tools;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Base64;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.stream.ImageOutputStream;

import loci.common.DebugTools;
import loci.formats.FormatException;
import loci.formats.ImageReader;
import loci.formats.gui.BufferedImageReader;

public class BioFormatsImageThumbnail {

    protected static final float JPEG_QUORITY = 0.85f;
    protected static final int BACKGROUND_COLOR = 0x000000;
    protected static final double MAXIMUM_SCALE = 2.0;

    public static BufferedImage resize(final BufferedImage image, final int width, final int height) {
        final boolean hasAlpha = image.getColorModel().hasAlpha();
        final int imageWidth = image.getWidth();
        final int imageHeight = image.getHeight();
        final double scaleWidth = (double) width / (double) imageWidth;
        final double scaleHeight = (double) height / (double) imageHeight;
        double scale = Math.min(scaleWidth, scaleHeight);
        if (scale > MAXIMUM_SCALE) {
            // limit scalling size to maximum scale
            scale = MAXIMUM_SCALE;
        }
        int resizeWidth = imageWidth;
        int resizeHeight = imageHeight;
        Image resizeImage = image;
        if (scale != 1.0) {
            // resize image if dimension is different with requrested dimension.
            resizeWidth = (int) (scale * (double) imageWidth);
            resizeHeight = (int) (scale * (double) imageHeight);
            resizeImage = image.getScaledInstance(resizeWidth, resizeHeight, Image.SCALE_AREA_AVERAGING);
        }
        final int type = hasAlpha ? BufferedImage.TYPE_INT_ARGB : BufferedImage.TYPE_INT_RGB;
        final BufferedImage ret = new BufferedImage(width, height, type);
        final Graphics2D g = ret.createGraphics();
        final Color c = new Color(image.getRGB(0, 0), hasAlpha);
        final int offsetX = (width - resizeWidth) / 2;
        final int offsetY = (height - resizeHeight) / 2;
        // System.out.println("original:" + imageWidth + "," + imageHeight);
        // System.out.println("resize:" + resizeWidth + "," + resizeHeight);
        // System.out.println("target:" + width + "," + height);
        // System.out.println("offset:" + offsetX + "," + offsetY);
        g.setColor(c);
        g.fillRect(0, 0, width, height);
        g.drawImage(resizeImage, offsetX, offsetY, resizeWidth, resizeHeight, null);
        g.dispose();

        return ret;
    }

    protected static BufferedImage removeAlphaChannel(final BufferedImage image, final int matColor) {
        if (!image.getColorModel().hasAlpha()) {
            return image;
        }
        final BufferedImage ret = new BufferedImage(image.getWidth(), image.getHeight(), BufferedImage.TYPE_INT_RGB);
        final Graphics2D g = ret.createGraphics();
        g.setColor(new Color(matColor, false));
        g.fillRect(0, 0, image.getWidth(), image.getHeight());
        g.drawImage(image, 0, 0, null);
        g.dispose();

        return ret;
    }

    protected static byte[] getJpegByteArray(final BufferedImage image, final float quality, final int matColor)
            throws BioFormatsImageException {
        byte[] ret = null;
        try (final ByteArrayOutputStream baos = new ByteArrayOutputStream();
                final ImageOutputStream ios = ImageIO.createImageOutputStream(baos);) {
            final ImageWriter writer = ImageIO.getImageWritersByFormatName("jpeg").next();
            final ImageWriteParam param = writer.getDefaultWriteParam();
            param.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
            param.setCompressionQuality(quality);
            writer.setOutput(ios);
            writer.write(null, new IIOImage(removeAlphaChannel(image, matColor), null, null), param);
            writer.dispose();
            ret = baos.toByteArray();
        } catch (final IOException e) {
            throw new BioFormatsImageException(e);
        }

        return ret;
    }

    protected static byte[] getPngByteArray(final BufferedImage image) throws BioFormatsImageException {
        byte[] ret = null;
        try (ByteArrayOutputStream baos = new ByteArrayOutputStream();
                BufferedOutputStream bos = new BufferedOutputStream(baos);) {
            image.flush();
            ImageIO.write(image, "png", bos);
            bos.flush();
            bos.close();
            ret = baos.toByteArray();
        } catch (final IOException e) {
            throw new BioFormatsImageException(e);
        }

        return ret;
    }

    public static String getDataUri(final BufferedImage image, final String mimeType) throws BioFormatsImageException {
        byte[] bytes = null;
        switch (mimeType) {
            case "image/png":
                bytes = getPngByteArray(image);
                break;
            case "image/jpeg":
                bytes = getJpegByteArray(image, JPEG_QUORITY, BACKGROUND_COLOR);
                break;
            default:
                throw new BioFormatsImageException("Unsupported image format: " + mimeType);
        }
        final String dataUri = "data:" + mimeType + ";base64," + Base64.getEncoder().encodeToString(bytes);

        return dataUri;
    }

    public static BufferedImage getImageFromBioFormatsFile(final String path) throws BioFormatsImageException {
        BufferedImage ret = null;
        final BufferedImageReader reader = new BufferedImageReader(new ImageReader());
        reader.setFlattenedResolutions(false);
        reader.setNormalized(true);
        reader.setOriginalMetadataPopulated(false);
        reader.setMetadataFiltered(true);
        try {
            reader.setId(path);
        } catch (FormatException | IOException e) {
            throw new BioFormatsImageException("unsupported file: " + path);
        }
        try {
            int series = 0;
            final int seriesCount = reader.getSeriesCount();
            for (int i = 0; i < seriesCount; i++) {
                reader.setSeries(i);
                if (reader.isThumbnailSeries()) {
                    series = i;
                    break;
                }
            }
            reader.setSeries(series);
            final int no = 0;
            // System.out.println("series count: " + seriesCount);
            // System.out.println("image count: " + reader.getImageCount());
            // System.out.println("resolution: " + reader.getResolutionCount());
            // System.out.println("thubmail series: " + series);
            ret = reader.openThumbImage(no);
        } catch (final Throwable e) {
            for (Throwable t = e; t != null; t = t.getCause()) {
                if (t instanceof OutOfMemoryError) {
                    throw new BioFormatsImageException("out of memory");
                }
            }
            throw new BioFormatsImageException(e);
        } finally {
            try {
                reader.close();
            } catch (final IOException e) {
                throw new BioFormatsImageException(e);
            }
        }

        return ret;
    }

    public static BufferedImage getImageFromFile(final String path) throws BioFormatsImageException {
        BufferedImage ret = null;
        final File file = new File(path);
        try {
            ret = ImageIO.read(file);
        } catch (final IOException e) {
            throw new BioFormatsImageException(e);
        }

        return ret;
    }

    public static void saveFile(final String dataUri, final String path) throws BioFormatsImageException {
        final Pattern pattern = Pattern.compile("^data:(image/[^;]+);(base64|utf8),(.*)$", Pattern.MULTILINE);
        final Matcher matcher = pattern.matcher(dataUri);
        if (!matcher.find()) {
            throw new BioFormatsImageException("Unsupported data URI format");
        }
        final byte[] bytes = matcher.group(2) == "utf8" ? matcher.group(3).getBytes()
                : Base64.getDecoder().decode(matcher.group(3));
        try (final FileOutputStream fos = new FileOutputStream(path);) {
            fos.write(bytes);
        } catch (final IOException e) {
            throw new BioFormatsImageException(e);
        }
    }

    public static boolean readThumbnail(final String path, final Map<String, Object> thumbnail) {
        boolean ret = true;
        try {
            final BufferedImage image = getImageFromBioFormatsFile(path);
            final int width = image.getWidth();
            final int height = image.getHeight();
            final String mimeType = image.getColorModel().hasAlpha() ? "image/png" : "image/jpeg";
            final String dataUri = getDataUri(image, mimeType);
            thumbnail.put("width", width);
            thumbnail.put("height", height);
            thumbnail.put("mimeType", mimeType);
            thumbnail.put("url", dataUri);
        } catch (final BioFormatsImageException e) {
            // ignore
            ret = false;
        }

        return ret;
    }

    public static void main(final String[] args) throws BioFormatsImageException {
        if (args.length != 1) {
            System.out.println("filename argument required");
            System.exit(1);
        }
        final String path = args[0];
        try {
            DebugTools.enableLogging("ERROR");
            BufferedImage image = getImageFromBioFormatsFile(path);
            final String format = image.getColorModel().hasAlpha() ? "png" : "jpeg";
            // image = resize(image, 160, 160);
            final String dataUri = getDataUri(image, "image/" + format);
            System.out.println(dataUri);
            // saveFile(dataUri, "/var/tmp/thumbnail." + format);
        } catch (final BioFormatsImageException e) {
            System.out.println(e.getMessage());
            System.exit(1);
        }
    }
}
